% Michael Vodhanel

% Purpose: Finds the independent components of a mixture set of observables

% Input:
%	X: The matrix observables where each row vector is a single observable 

% Output:
%	U: The Independent Components where each row vector is an independent component

function U = ICA(X)

N = size(X, 1);
W = zeros(N);

% Preprocess (whiten) data
[X, m, E, D] = Preprocess(X);

% Perform the ICA iteration
count = 0;
w = ones(N, 1);
w = w / norm(w);
%w_old = zeros(size(w));
W = repmat(w, 1, N);
W = 1 + rand(N);
while ( count < 100 ) %(norm(w - w_old) > .001) & (norm(w + w_old) > .001) & (count < 100) ) 
    for n = 1:N 
        lump=W(:,n)'*X;
        glump=lump.*exp(-lump.^2/2);   
        gplump=(1-lump.^2).*exp(-lump.^2/2);

        W(:,n) =X*glump'-mean(gplump)*W(:,n);
        
        % Gram Schmidt decorrelation of nth vector with the previous n-1
        Decor = zeros(size(w));
        for j=1:n-1
           Decor = W(:,j) * dot(W(:,n),W(:,j)) / dot(W(:,j),W(:,j));
           W(:,n) = W(:,n) - Decor;
        end
                
        W(:,n) = W(:,n) / norm(W(:,n));
    end
    count = count + 1;
end
W = W';

for i = 1:N
   W(i,:) = W(i,:) / sum(abs(W(i,:)));
   W(i,:) = W(i,:) / sum(W(i,:)); 
end

% Find the seperated signals
U = W*X;

% Find important matrices
%W = W
Awig = inv(W);
A = inv(E*inv(sqrt(D))*E') * Awig;

% Add means back
m = W*m;
for i = 1:N
    U(i,:) = U(i,:) + m(i);
end

% Use reconstructed A matrix to get proper signs back on U
% This works if A is assumed to be all positive to begin with.
%U = sign(diag(sum(A))) * U;

% Unpreprocess the signals
%U = Postprocess(X, W, m, UnWhitener);
%U = W*X;
%W = W * UnWhitener
